package ru.dbotthepony.kommons.arrays

fun interface ObjectMatrixInitializer<out T> {
	fun invoke(column: Int, row: Int): T
}

private fun initialize(columns: Int, rows: Int, initializer: ObjectMatrixInitializer<Any?>): Array<Any?> {
	require(columns >= 0 && rows >= 0) { "Invalid dimensions: $columns x $rows" }

	var c = 0
	var r = 0

	return Array(columns * rows) {
		val value = initializer.invoke(c, r++)

		if (r == rows) {
			c++
			r = 0
		}

		value
	}
}

private fun initialize(columns: Int, rows: Int, value: Any?): Array<Any?> {
	require(columns >= 0 && rows >= 0) { "Invalid dimensions: $columns x $rows" }
	return Array(columns * rows) { value }
}

class Object2DArray<T> private constructor(
	override val columns: Int,
	override val rows: Int,
	private val mem: Array<Any?>
) : Array2D() {
	constructor(
		columns: Int,
		rows: Int,
		initializer: ObjectMatrixInitializer<T>
	) : this(columns, rows, initialize(columns, rows, initializer))

	constructor(
		columns: Int,
		rows: Int,
		value: T
	) : this(columns, rows, initialize(columns, rows, value))

	constructor(
		parent: Object2DArray<T>
	) : this(parent.columns, parent.rows, Array(parent.size) { parent.mem[it] })

	init {
		require(columns >= 0 && rows >= 0) { "Invalid dimensions: $columns x $rows" }
	}

	override fun equals(other: Any?): Boolean {
		return matrixEquals(this, other) { a, b, c, d -> a[c, d] == b[c, d] }
	}

	override fun hashCode(): Int {
		return matrixHashCode(this) { a, b, c -> a[b, c].hashCode() }
	}

	override fun toString(): String {
		return matrixToString(this) { a, b, c -> a[b, c].toString() }
	}

	fun store(into: Object2DArray<T>) {
		into.load(this)
	}

	fun load(other: Object2DArray<T>) {
		requireSizeEquals(other)

		for (column in columnIndices) {
			for (row in rowIndices) {
				this[column, row] = other[column, row]
			}
		}
	}

	operator fun get(column: Int, row: Int): T {
		try {
			return mem[column * rows + row] as T
		} catch (err: IndexOutOfBoundsException) {
			throw IndexOutOfBoundsException("Matrix index out of bounds: $column x $row, while this array has dimensions of $columns x $rows")
		}
	}

	operator fun set(column: Int, row: Int, value: T) {
		try {
			mem[column * rows + row] = value
		} catch (err: IndexOutOfBoundsException) {
			throw IndexOutOfBoundsException("Matrix index out of bounds: $column x $row, while this array has dimensions of $columns x $rows")
		}
	}

	val transposed get(): Object2DArray<T> {
		val new = nulls<T>(rows, columns)

		for (column in columnIndices) {
			for (row in rowIndices) {
				new[column, row] = this[row, column]
			}
		}

		return new as Object2DArray<T>
	}

	companion object {
		@JvmStatic
		fun <T> nulls(columns: Int, rows: Int): Object2DArray<T?> {
			require(columns >= 0 && rows >= 0) { "Invalid dimensions: $columns x $rows" }
			return Object2DArray(columns, rows, arrayOfNulls(columns * rows))
		}

		@JvmField
		val EMPTY = nulls<Any?>(0, 0)

		@JvmStatic
		fun <T> empty(): Object2DArray<T> {
			return EMPTY as Object2DArray<T>
		}
	}
}
